import torch
import torch.nn as nn
from typing import List
import torch.nn.functional as F
import segmentation_models_pytorch as smp

def load_model(model_name:str, input_bands, output_bands,corrector:bool=True, downscale_factor_corrector:int=1,
               backbone_unet:str="mobilenet_v2") -> nn.Module:
    input_channel_size = len(input_bands)
    output_channel_size = len(output_bands)

    if model_name == 'SimpleCNN':
        model = SimpleCNN(input_channel_size=input_channel_size, output_channel_size=output_channel_size)
    elif model_name == 'SimpleCNNAE':
        assert input_channel_size == output_channel_size
        model = SimpleConvolutionalAutoencoder(channel_size=input_channel_size)
    elif model_name == 'SimpleAE':
        assert input_channel_size == output_channel_size
        model = SimpleAutoencoder(channel_size=input_channel_size)
    elif model_name == 'CNNAE':
        assert input_channel_size == output_channel_size
        model = ConvolutionalAutoencoder(channel_size=input_channel_size)
    elif model_name == 'CNN':
        model = CNN(channel_size=input_channel_size, output_channel_size=output_channel_size)
    elif model_name == 'ComplexCNN_small':
        assert input_channel_size == output_channel_size
        model = ComplexCNN_small(channel_size=input_channel_size)
    elif model_name == 'ComplexCNN_medium':
        assert input_channel_size == output_channel_size
        model = ComplexCNN_medium(channel_size=input_channel_size)
    elif model_name == 'ComplexCNN_big':
        assert input_channel_size == output_channel_size
        model = ComplexCNN_big(channel_size=input_channel_size)
    elif model_name == 'CNNAE_2':
        assert input_channel_size == output_channel_size
        model = ConvolutionalAutoencoder_1(channel_size=input_channel_size)
    elif model_name == 'CNNAE_2_w_dropout':
        assert input_channel_size == output_channel_size
        model = ConvolutionalAutoencoder_1_w_dropout(channel_size=input_channel_size, dropout=0.2)
    elif model_name == 'CNNAE_3':
        assert input_channel_size == output_channel_size
        model = ConvolutionalAutoencoder_2(channel_size=input_channel_size)
    elif model_name == 'SimpleRSNet':
        model = SimpleRSNet(input_channel_size, output_channel_size)
    elif model_name == 'Unet':
        model = smp.Unet(encoder_name=backbone_unet, encoder_weights=None, in_channels=input_channel_size,
                         classes=output_channel_size)
    elif model_name == 'DeepLabV3':
        # 'timm-mobilenetv3_large_100'
        model = smp.DeepLabV3(encoder_name=backbone_unet, encoder_weights=None,
                              in_channels=input_channel_size, classes=output_channel_size)
    else:
        raise NotImplementedError(f"Model {model_name} not implemented")

    if corrector:
        channels_input_in_output = [input_bands.index(b) for b in output_bands]

        model = Corrector(model, channels_input_in_output=channels_input_in_output,
                          downscale_factor=downscale_factor_corrector)

    return model


def load_cloud_detection_model():
    return smp.Unet(
        encoder_name="mobilenet_v2",
        encoder_weights=None,
        in_channels=13,
        classes=4,
    )

class Corrector(nn.Module):
    def __init__(self, model:nn.Module, channels_input_in_output:List[int], downscale_factor:int=1):
        super(Corrector,self).__init__()
        self.model = model
        self.channels_input_in_output = channels_input_in_output
        self.downscale_factor = downscale_factor

    def forward(self, x):
        if self.downscale_factor > 1:
            x_model = F.interpolate(x, scale_factor=1 / self.downscale_factor, mode="bilinear", antialias=True)
        else:
            x_model = x

        correction = self.model(x_model)
        if self.downscale_factor > 1:
            correction = F.interpolate(correction, size=x.shape[-2:], mode="bilinear")

        input_to_output = x[:, self.channels_input_in_output, ...]

        return input_to_output - correction


class SimpleCNN(nn.Module):
    def __init__(self,input_channel_size=13, output_channel_size=12):
        super(SimpleCNN,self).__init__()
        self.cnn_layers=torch.nn.Sequential(
                        torch.nn.Conv2d(in_channels=input_channel_size,
                                  out_channels=output_channel_size,
                                  kernel_size=1,
                                  stride=1))
    def forward(self, x):
        x = self.cnn_layers(x)
        return x

class CNN(nn.Module):
    def __init__(self,channel_size=3, output_channel_size=None):
        if output_channel_size is None:
            output_channel_size = channel_size

        super(CNN,self).__init__()
        self.cnn_layers=torch.nn.Sequential(
                        torch.nn.Conv2d(in_channels=channel_size,
                                        out_channels=int(channel_size*4),
                                        kernel_size=1,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*4),
                                        out_channels=output_channel_size,
                                        kernel_size=1,
                                        stride=1)
        )
    def forward(self,x):
        x = self.cnn_layers(x)
        return x

class ComplexCNN_small(nn.Module):
    def __init__(self,channel_size=3):
        super(ComplexCNN_small,self).__init__()
        self.cnn_layers=torch.nn.Sequential(
                        torch.nn.Conv2d(in_channels=channel_size,
                                        out_channels=int(channel_size*2),
                                        kernel_size=3,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*2),
                                        out_channels=int(channel_size*4),
                                        kernel_size=6,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*4),
                                        out_channels=int(channel_size*2),
                                        kernel_size=6,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*2),
                                        out_channels=channel_size,
                                        kernel_size=3,
                                        padding=7,
                                        stride=1)
        )
    def forward(self,x):
        x = self.cnn_layers(x)
        return x

class ComplexCNN_medium(nn.Module):
    def __init__(self,channel_size=3):
        super(ComplexCNN_medium,self).__init__()
        self.cnn_layers=torch.nn.Sequential(
                        torch.nn.Conv2d(in_channels=channel_size,
                                        out_channels=int(channel_size*4),
                                        kernel_size=3,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*4),
                                        out_channels=int(channel_size*8),
                                        kernel_size=7,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*8),
                                        out_channels=int(channel_size*4),
                                        kernel_size=7,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*4),
                                        out_channels=channel_size,
                                        kernel_size=3,
                                        padding=8,
                                        stride=1)
        )
    def forward(self,x):
        x = self.cnn_layers(x)
        return x

class ComplexCNN_big(nn.Module):
    def __init__(self,channel_size=3):
        super(ComplexCNN_big,self).__init__()
        self.cnn_layers=torch.nn.Sequential(
                        torch.nn.Conv2d(in_channels=channel_size,
                                        out_channels=int(channel_size*4),
                                        kernel_size=3,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*4),
                                        out_channels=int(channel_size*8),
                                        kernel_size=3,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*8),
                                        out_channels=int(channel_size*16),
                                        kernel_size=7,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*16),
                                        out_channels=int(channel_size*8),
                                        kernel_size=7,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*8),
                                        out_channels=int(channel_size*4),
                                        kernel_size=3,
                                        stride=1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv2d(in_channels=int(channel_size*4),
                                        out_channels=channel_size,
                                        kernel_size=3,
                                        padding=10,
                                        stride=1)
        )
    def forward(self,x):
        x = self.cnn_layers(x)
        return x

###### Linear Autoencoder ######
class SimpleAutoencoder(nn.Module):
    def __init__(self, image_shape=(509, 509), channel_size=3):
        self.image_shape=image_shape
        self.channel_size=channel_size
        super().__init__()
        self.encoder = torch.nn.Sequential(
            torch.nn.Linear(image_shape[0]*image_shape[1]*channel_size, 509),
            torch.nn.ReLU(),
            torch.nn.Linear(509, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 128),
            torch.nn.ReLU(),
            torch.nn.Linear(128, 64),
            torch.nn.ReLU(),
            torch.nn.Linear(64, 32)
        )

        self.decoder = torch.nn.Sequential(
            torch.nn.Linear(32, 64),
            torch.nn.ReLU(),
            torch.nn.Linear(64, 128),
            torch.nn.ReLU(),
            torch.nn.Linear(128, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 509),
            torch.nn.ReLU(),
            torch.nn.Linear(509, image_shape[0]*image_shape[1]*channel_size),
        )

    def forward(self, x):
        x=x.reshape(-1,self.image_shape[0]*self.image_shape[1]*self.channel_size)
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded.reshape(x.shape[0],self.channel_size,self.image_shape[0],self.image_shape[1])


###### Convolutional Autoencoders ######
class SimpleConvolutionalAutoencoder(nn.Module):
    def __init__(self, channel_size=3):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(in_channels=channel_size,
                      out_channels=64,
                      kernel_size=3,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(in_channels=64,
                      out_channels=32,
                      kernel_size=3,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(in_channels=32,
                      out_channels=16,
                      kernel_size=2,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True)
        )
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(in_channels=16,
                               out_channels=32,
                               kernel_size=2,
                               stride=1),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels=32,
                               out_channels=64,
                               kernel_size=2,
                               stride=1),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels=64,
                               out_channels=channel_size,
                               kernel_size=4,
                               stride=1)
        )
    def forward(self, x):
        x=self.encoder(x)
        x=self.decoder(x)
        return x

class ConvolutionalAutoencoder(nn.Module):
    def __init__(self, channel_size=3):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(in_channels=channel_size,
                      out_channels=128,
                      kernel_size=3,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True),
            nn.BatchNorm2d(128),
            nn.Conv2d(in_channels=128,
                      out_channels=64,
                      kernel_size=3,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True),
            nn.BatchNorm2d(64),
            nn.Conv2d(in_channels=64,
                      out_channels=32,
                      kernel_size=2,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True),
            nn.BatchNorm2d(32),
            nn.Conv2d(in_channels=32,
                      out_channels=16,
                      kernel_size=1,
                      stride=1),
            nn.LeakyReLU(0.2, inplace=True)
        )
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(in_channels=16,
                               out_channels=32,
                               kernel_size=2,
                               stride=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(32),
            nn.ConvTranspose2d(in_channels=32,
                               out_channels=64,
                               kernel_size=2,
                               stride=1),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(64),
            nn.ConvTranspose2d(in_channels=64,
                               out_channels=channel_size,
                               kernel_size=4,
                               stride=1)
        )
    def forward(self, x):
        x=self.encoder(x)
        x=self.decoder(x)
        return x

class ConvolutionalAutoencoder_1(nn.Module):
    def __init__(self, channel_size=3):
        super().__init__()
        self.encoder=nn.Sequential(
            nn.Conv2d(channel_size, 64, 3, stride=2, padding=1),
            nn.ReLU(),
            nn.Conv2d(64,128,3,stride=2,padding=1),
            nn.ReLU(),
            nn.Conv2d(128,256,7),
            nn.ReLU()
        )
        self.decoder=nn.Sequential(
            nn.ConvTranspose2d(256,128,7),
            nn.ReLU(),
            nn.ConvTranspose2d(128,64,3,stride=2,padding=1,output_padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(64,12,3,stride=2,padding=1,output_padding=1),
        )
    def forward(self,x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x

class ConvolutionalAutoencoder_1_w_dropout(nn.Module):
    def __init__(self, channel_size=3, dropout=0.2):
        super().__init__()
        self.encoder=nn.Sequential(
            nn.Conv2d(channel_size, 64, 3, stride=2, padding=1),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Conv2d(64,128,3,stride=2,padding=1),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Conv2d(128,256,7),
            nn.ReLU(),
            nn.Dropout(dropout),
        )
        self.decoder=nn.Sequential(
            nn.ConvTranspose2d(256,128,7),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.ConvTranspose2d(128,64,3,stride=2,padding=1,output_padding=1),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.ConvTranspose2d(64,12,3,stride=2,padding=1,output_padding=1),
        )
    def forward(self,x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x

class ConvolutionalAutoencoder_2(nn.Module):
    def __init__(self, channel_size=3):
        super().__init__()
        self.encoder=nn.Sequential(
            nn.Conv2d(channel_size, 128, 3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(128,256,5,stride=1,padding=1),
            nn.ReLU()
            )
        self.decoder=nn.Sequential(
            nn.ConvTranspose2d(256,128,5),
            nn.ReLU(),
            nn.ConvTranspose2d(128,12,3,stride=1,padding=2),
        )
    def forward(self,x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x



class DSConv2d(torch.nn.Module):
    def __init__(self, nin, nout, kernel_size = 3, padding = 1, bias=False):
        super(DSConv2d, self).__init__()
        self.depthwise = torch.nn.Conv2d(nin, nin, kernel_size=kernel_size, padding=padding, groups=nin, bias=bias)
        self.pointwise = torch.nn.Conv2d(nin, nout, kernel_size=1, bias=bias)

    def forward(self, x):
        out = self.depthwise(x)
        out = self.pointwise(out)
        return out

def SimpleRSNet(num_channels=13, out_channels=12):
    return torch.nn.Sequential(
        DSConv2d(num_channels, 64, kernel_size=5, padding=2),
        torch.nn.ReLU(inplace=True),
        DSConv2d(64, 64, kernel_size=3, padding=1),
        torch.nn.ReLU(inplace=True),
        DSConv2d(64, 32, kernel_size=3, padding=1),
        torch.nn.ReLU(inplace=True),
        DSConv2d(32, out_channels, kernel_size=3, padding=1)
    )

class SimpleRSNetPredict(torch.nn.Module):
        def __init__(self, num_channels=13, out_channels=12):
            super().__init__()
            self.model = SimpleRSNet(num_channels, out_channels)
        def forward(self, x):
            return self.model(x)
        



import math
from typing import Tuple, List, Optional, Callable
import torch
import segmentation_models_pytorch as smp
import numpy as np
from dtacs import nn


def find_padding(v:int, divisor=32) -> Tuple[Tuple[int, int],slice]:
    v_divisible = max(divisor, int(divisor * math.ceil(v / divisor)))
    total_pad = v_divisible - v
    pad_1 = total_pad // 2
    pad_2 = total_pad - pad_1
    slice_rows = slice(pad_1, None if pad_2 <= 0 else -pad_2)
    return (pad_1, pad_2), slice_rows


class ACModel:
    def __init__(self, input_bands:List[str], output_bands:List[str], device:torch.device=torch.device("cpu"),):
        super().__init__()

        self.model = nn.load_model("CNN",
                                   input_bands, output_bands,
                                   corrector=True)
        self.device = device
        self.model.eval()
        self.model.to(self.device)

    def load_weights(self, path:str):
        with open(path,"rb") as fh:
            weights =  torch.load(fh, map_location=self.device)
        self.model.load_state_dict(weights["state_dict"])

    def predict(self, tensor: np.array) -> np.array:
        """
            tensor: np.array (`in_channels`, H, W) with reflectances multiplied by 10000

        Returns:
            np.uint16 np.array (H, W) with surface reflectances multiplied by 10000
        """

        pad_r,slice_rows = find_padding(tensor.shape[-2])
        pad_c, slice_cols= find_padding(tensor.shape[-1])

        tensor_padded = np.pad(
            tensor, ((0, 0), (pad_r[0], pad_r[1]), (pad_c[0], pad_c[1])), "reflect"
        )

        tensor_padded /= 10_000  # assume values are as downloaded
        tensor_padded = torch.tensor(tensor_padded, device=self.device)[None]  # Add batch dim

        with torch.no_grad():
            pred_padded = self.model(tensor_padded)[0]
            pred_cont = torch.clamp(pred_padded[(slice(None), slice_rows, slice_cols)], 0, 2)
            pred_cont *= 10_000

        return np.array(pred_cont.cpu()).astype(np.uint16)


class CDModel(torch.nn.Module):
    def __init__(self, device:torch.device=torch.device("cpu") ,model:Optional[Callable]=None, in_channels:int=13):
        super().__init__()
        if model is None:
            self.model = smp.Unet(
                encoder_name="mobilenet_v2",
                encoder_weights=None,
                in_channels=in_channels,
                classes=4
            )
        else:
            self.model = model
        self.in_channels = in_channels
        self.device = device
        self.model.eval()
        self.model.to(self.device)

    def load_weights(self, path:str):
        with open(path, "rb") as fh:
            weights =  torch.load(fh, map_location=self.device)
        self.load_state_dict(weights["state_dict"])

    def predict(self, tensor: np.array) -> np.array:
        """
            tensor: np.array (`in_channels`, H, W) with reflectances multiplied by 10000

        Returns:
            np.uint8 np.array (H, W) with interpretation {0: clear, 1: Thick cloud, 2: thin cloud, 3: cloud shadow}
        """

        pad_r,slice_rows = find_padding(tensor.shape[-2])
        pad_c, slice_cols= find_padding(tensor.shape[-1])

        tensor_padded = np.pad(
            tensor, ((0, 0), (pad_r[0], pad_r[1]), (pad_c[0], pad_c[1])), "reflect"
        )

        tensor_padded /= 10_000  # assume values are as downloaded
        tensor_padded = torch.tensor(tensor_padded, device=self.device)[None]  # Add batch dim

        with torch.no_grad():
            pred_padded = self.model(tensor_padded)[0]
            pred_cont = pred_padded[(slice(None), slice_rows, slice_cols)]
            pred_discrete = torch.argmax(pred_cont, dim=0).type(torch.uint8)

        return np.array(pred_discrete.cpu())


from matplotlib import colors
import matplotlib.patches as mpatches
import numpy as np
import matplotlib.pyplot as plt


COLORS_CLOUDSEN12 = np.array([[139, 64, 0], # clear
                              [220, 220, 220], # Thick cloud
                              [180, 180, 180], # Thin cloud
                              [60, 60, 60]], # cloud shadow
                             dtype=np.float32) / 255

INTERPRETATION_CLOUDSEN12 = ["clear", "Thick cloud", "Thin cloud", "Cloud shadow"]

INTERPRETATION_DYNAMIC_WORLD = ["water", "trees", "grass", "flooded_vegetation",  "crops", "shrub_and_scrub", "built", "bare", "snow_and_ice"]
COLORS_DYNAMIC_WORLD_HEX = ["0000FF","007700", "88B053","42f5e0", "E49635", "DFC35A", "C4281B","A59B8F","B39FE1"]
COLORS_DYNAMIC_WORLD_CLOUDS_HEX = ["0000FF","007700", "88B053","42f5e0", "E49635", "DFC35A", "C4281B","A59B8F","B39FE1", "000000"]
COLORS_DYNAMIC_WORLD = np.array([tuple(int(c[i:i+2], 16) for i in (0, 2, 4)) for c in COLORS_DYNAMIC_WORLD_HEX])/255
COLORS_DYNAMIC_WORLD_CLOUDS = np.array([tuple(int(c[i:i+2], 16) for i in (0, 2, 4)) for c in COLORS_DYNAMIC_WORLD_CLOUDS_HEX])/255
INTERPRETATION_DYNAMIC_WORLD_CLOUDS = INTERPRETATION_DYNAMIC_WORLD + ["cloud"]

def plot_segmentation_mask(mask, color_array, interpretation_array=None,legend:bool=True, ax=None):
    """
    Args:
        mask: (H, W) np.array
        color_array: colors for values 0,...,len(color_array)-1 of mask
        interpretation_array: interpretation for classes 0, ..., len(color_array)-1
        
    """
    cmap_categorical = colors.ListedColormap(color_array)
    
    norm_categorical = colors.Normalize(vmin=-.5,
                                        vmax=color_array.shape[0]-.5)
    
    color_array = np.array(color_array)
    if interpretation_array is not None:
        assert len(interpretation_array) == color_array.shape[0], f"Different numbers of colors and interpretation {len(interpretation_array)} {color_array.shape[0]}"
    
    
    if ax is None:
        ax = plt.gca()
    
    ax.imshow(mask, cmap=cmap_categorical, norm=norm_categorical,interpolation='nearest')
    if legend:
        patches = []
        for c, interp in zip(color_array, interpretation_array):
            patches.append(mpatches.Patch(color=c, label=interp))
        
        ax.legend(handles=patches,
                  loc='upper right')
    return ax


def plot_cloudSEN12mask(mask, legend:bool=True, ax=None):
    """
    Args:
        mask: (H, W) np.array        
    """
    
    return plot_segmentation_mask(mask=mask,color_array=COLORS_CLOUDSEN12,
                                  interpretation_array=INTERPRETATION_CLOUDSEN12,legend=legend,ax=ax)


def plot_dynamicearthnet(mask, legend:bool=True, ax=None):
    
    return plot_segmentation_mask(mask,color_array=COLORS_DYNAMIC_WORLD,
                                  interpretation_array=INTERPRETATION_DYNAMIC_WORLD_CLOUDS,ax=ax,
                                  legend=legend)
